@@ -593,11 +593,11 @@ gpgtar_list (const char *filename, int decrypt)
     }
 
  leave:
-  /* fixme: Why can't we use gpgrt_process_release (proc); */
   free_strlist (extheader);
   xfree (header);
   if (stream != es_stdin)
     es_fclose (stream);
+  gpgrt_process_release (proc);
   return err;
 }
 
